from __future__ import print_function

import sys

from tox.session import prepare as tox_prepare

from detox import __version__
from detox.proc import Detox


def main(args=None):
    args = sys.argv[1:] if args is None else args
    if args and args[0] == "--version":
        print("detox {} running as plugin in tox:".format(__version__))
        # fall through to let tox add its own version info ...
    config = tox_prepare(args)
    detox = Detox(config)
    if not hasattr(config.option, "quiet_level") or not config.option.quiet_level:
        detox.startloopreport()
    ret = detox.runtestsmulti(detox.toxsession.evaluated_env_list())
    print("### WARNING ###\n\n"
          "detox is not compatible with versions of tox > 3.6."
          "Consider uninstalling detox and upgrading tox to >= 3.7"
          "to use its parallel mode.")
    return ret



import argparse
import multiprocessing

from tox import hookimpl


@hookimpl
def tox_addoption(parser):
    def positive_integer(value):
        ivalue = int(value)
        if ivalue <= 0:
            raise argparse.ArgumentTypeError("{} must be greater 0".format(value))
        return ivalue

    try:
        num_proc = multiprocessing.cpu_count()
    except Exception:
        num_proc = 2
    parser.add_argument(
        "-n",
        "--num",
        type=positive_integer,
        action="store",
        default=num_proc,
        dest="numproc",
        help="set the number of concurrent processes " "(default {}).".format(num_proc),
    )



# -*- coding: utf-8 -*-


import subprocess
import sys


def invoke(command, success_codes=(0,)):
    try:
        output = subprocess.check_output(command, stderr=subprocess.STDOUT)
        status = 0
    except subprocess.CalledProcessError as error:
        output = error.output
        status = error.returncode
    output = output.decode("utf-8")
    if status not in success_codes:
        raise Exception(
            'Command %r return exit code %d and output: """%s""".'
            % (command, status, output)
        )
    return status, output


def test_run_as_module():
    """Can be run as `python -m detox ...`."""
    status, output = invoke([sys.executable, "-m", "detox", "--help"])
    assert status == 0
    assert output.startswith("usage:")



import io
from setuptools import setup


def make_long_description():
    with io.open("README.md", encoding='UTF-8') as f:
        readme = f.read()
    with io.open("CHANGELOG", encoding='UTF-8') as f:
        changelog = f.read()
    return readme + "\n\n" + changelog


setup(
    name="detox",
    description="distributing activities of the tox tool",
    long_description=make_long_description(),
    long_description_content_type="text/markdown",
    version="0.19",  # Note: keep in sync with detox/__init__.py
    url="https://github.com/tox-dev/detox",
    license="MIT",
    platforms=["unix", "linux", "osx", "cygwin", "win32"],
    author="holger krekel",
    author_email="holger@merlinux.eu",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Framework :: tox",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: MacOS :: MacOS X",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Software Development :: Testing",
        "Topic :: Software Development :: Libraries",
        "Topic :: Utilities",
        "Programming Language :: Python",
    ],
    packages=["detox"],
    install_requires=["tox>=3.5,<3.7", "py>=1.4.27", "eventlet>=0.15.0"],
    extras_require={"lint": ["black", "flake8"], "dev": ["pytest >= 3.8"]},
    entry_points={
        "console_scripts": "detox=detox.cli:main",
        "tox": ["proclimit = detox.tox_proclimit"],
    },
)



from __future__ import with_statement

import eventlet
from eventlet.timeout import Timeout
from eventlet.green.subprocess import Popen
from eventlet import GreenPool
import tox
import tox.session


def timelimited(secs, func):
    if secs is not None:
        with Timeout(secs):
            return func()
    return func()


class FileSpinner:
    chars = r"- \ | / - \ | /".split()

    def __init__(self):
        self.path2last = {}

    def getchar(self, path):
        try:
            lastsize, charindex = self.path2last[path]
        except KeyError:
            lastsize, charindex = 0, 0
        newsize = 0 if not path else path.size()
        if newsize != lastsize:
            charindex += 1
        self.path2last[path] = (lastsize, charindex)
        return self.chars[charindex % len(self.chars)]


class ToxReporter(tox.session.Reporter):
    sortorder = (
        "runtests command installdeps installpkg inst inst-nodeps "
        "sdist-make create recreate".split()
    )

    def __init__(self, session):
        super(ToxReporter, self).__init__(session)
        self._actionmayfinish = set()

    def _loopreport(self):
        filespinner = FileSpinner()
        while 1:
            eventlet.sleep(0.2)
            msg = []
            ac2popenlist = {}
            for action in self.session._actions:
                for popen in action._popenlist:
                    if popen.poll() is None:
                        ol = ac2popenlist.setdefault(action.activity, [])
                        ol.append(popen)
                if not action._popenlist and action in self._actionmayfinish:
                    super(ToxReporter, self).logaction_finish(action)
                    self._actionmayfinish.remove(action)

            for acname in self.sortorder:
                try:
                    popenlist = ac2popenlist.pop(acname)
                except KeyError:
                    continue
                sublist = []
                for popen in popenlist:
                    name = getattr(popen.action.venv, "name", "INLINE")
                    char = filespinner.getchar(popen.outpath)
                    sublist.append("%s%s" % (name, char))
                msg.append("%s %s" % (acname, " ".join(sublist)))
            assert not ac2popenlist, ac2popenlist
            if msg:
                msg = "   ".join(msg)
                if len(msg) >= self.tw.fullwidth:
                    msg = msg[: self.tw.fullwidth - 3] + ".."
                self.tw.reline(msg)

    def __getattr__(self, name):
        if name[0] == "_":
            raise AttributeError(name)

        def generic_report(*args):
            self._calls.append((name,) + args)
            if self.config.option.verbosity >= 2:
                print("%s" % (self._calls[-1],))

        return generic_report

    def logaction_finish(self, action):
        if action._popenlist:
            # defer finishing output to spinner loop
            self._actionmayfinish.add(action)
        else:
            super(ToxReporter, self).logaction_finish(action)


class Detox:
    def __init__(self, toxconfig):
        self._toxconfig = toxconfig
        self._resources = Resources(self)
        self._sdistpath = None
        self._toxsession = None

    def startloopreport(self):
        if self.toxsession.report.tw.hasmarkup:
            eventlet.spawn_n(self.toxsession.report._loopreport)

    @property
    def toxsession(self):
        if not self._toxsession:
            self._toxsession = tox.session.Session(
                self._toxconfig, Report=ToxReporter, popen=Popen
            )
        return self._toxsession

    def provide_sdist(self):
        try:
            sdistpath = self.toxsession.get_installpkg_path()  # tox < 3.3
        except AttributeError:
            from tox.package import get_package

            sdistpath = get_package(self.toxsession)
        if not sdistpath:
            raise SystemExit(1)
        return sdistpath

    def provide_venv(self, venvname):
        venv = self.toxsession.getvenv(venvname)
        if self.toxsession.setupenv(venv):
            return venv

    def provide_installpkg(self, venvname, sdistpath):
        venv = self.toxsession.getvenv(venvname)
        return self.toxsession.installpkg(venv, sdistpath)

    def provide_developpkg(self, venvname):
        venv = self.toxsession.getvenv(venvname)
        return self.toxsession.developpkg(venv, self.toxsession.config.setupdir)

    def runtests(self, venvname):
        if self.toxsession.config.option.sdistonly:
            self._sdistpath = self.getresources("sdist")
            return
        if self.toxsession.config.skipsdist:
            venv, = self.getresources("venv:%s" % venvname)
            if venv:
                venv.finish()
                self.toxsession.runtestenv(venv, redirect=True)
        else:
            venv, sdist = self.getresources("venv:%s" % venvname, "sdist")
            if venv and sdist:
                # tox >= 3.5 returns a tuple, the first one is the session package
                if isinstance(sdist, tuple):
                    sdist = sdist[0]
                if self.toxsession.installpkg(venv, sdist):
                    self.toxsession.runtestenv(venv, redirect=True)

    def runtestsmulti(self, envlist):
        pool = GreenPool(size=self._toxconfig.option.numproc)
        threads = []
        for env in envlist:
            threads.append(pool.spawn(self.runtests, env))

        for t in threads:
            # re-raises any exceptions of the worker thread
            t.wait()
        if not self.toxsession.config.option.sdistonly:
            retcode = self._toxsession._summary()
            return retcode

    def getresources(self, *specs):
        return self._resources.getresources(*specs)


class Resources:
    def __init__(self, providerbase):
        self._providerbase = providerbase
        self._spec2thread = {}
        self._pool = GreenPool()
        self._resources = {}

    def _dispatchprovider(self, spec):
        parts = spec.split(":")
        name = parts.pop(0)
        provider = getattr(self._providerbase, "provide_" + name)
        self._resources[spec] = res = provider(*parts)
        return res

    def getresources(self, *specs):
        for spec in specs:
            if spec not in self._resources:
                if spec not in self._spec2thread:
                    t = self._pool.spawn(self._dispatchprovider, spec)
                    self._spec2thread[spec] = t
        resources = []
        for spec in specs:
            if spec not in self._resources:
                self._spec2thread[spec].wait()
            resources.append(self._resources[spec])
        return resources



from datetime import datetime, timedelta

import pytest
import eventlet

from detox.proc import Resources


class TestResources:
    def test_getresources(self):
        x = []

        class Provider:
            def provide_abc(self):
                x.append(1)
                return 42

        resources = Resources(Provider())
        res, = resources.getresources("abc")
        assert res == 42
        assert len(x) == 1
        res, = resources.getresources("abc")
        assert len(x) == 1
        assert res == 42

    def test_getresources_param(self):
        class Provider:
            def provide_abc(self, param):
                return param

        resources = Resources(Provider())
        res, = resources.getresources("abc:123")
        return res == "123"

    def test_getresources_parallel(self):
        x = []

        class Provider:
            def provide_abc(self):
                x.append(1)
                return 42

        resources = Resources(Provider())
        pool = eventlet.GreenPool(2)
        pool.spawn(lambda: resources.getresources("abc"))
        pool.spawn(lambda: resources.getresources("abc"))
        pool.waitall()
        assert len(x) == 1

    def test_getresources_multi(self):
        x = []

        class Provider:
            def provide_abc(self):
                x.append(1)
                return 42

            def provide_def(self):
                x.append(1)
                return 23

        resources = Resources(Provider())
        a, d = resources.getresources("abc", "def")
        assert a == 42
        assert d == 23


class TestDetoxExample1:
    pytestmark = [pytest.mark.example1, pytest.mark.timeout(20)]

    def test_createsdist(self, detox):
        sdists, = detox.getresources("sdist")
        for sdist in sdists:
            assert sdist.check()

    def test_getvenv(self, detox):
        venv, = detox.getresources("venv:py")
        assert venv.envconfig.envdir.check()
        venv2, = detox.getresources("venv:py")
        assert venv == venv2

    def test_test(self, detox):
        detox.runtests("py")


class TestDetoxExample2:
    pytestmark = [pytest.mark.example2, pytest.mark.timeout(20)]

    def test_test(self, detox):
        detox.runtests("py")

    def test_developpkg(self, detox):
        detox.getresources("venv:py")
        developpkg, = detox.getresources("developpkg:py")
        assert developpkg is False


class TestCmdline:
    pytestmark = [pytest.mark.example1]

    @pytest.mark.timeout(20)
    def test_runtests(self, cmd):
        result = cmd.rundetox("-e", "py", "-v", "-v")
        result.stdout.fnmatch_lines(["py*getenv*", "py*create:*"])


class TestProcLimitOption:
    pytestmark = [pytest.mark.example3]

    def test_runtestmulti(self):
        class MyConfig:
            class MyOption:
                numproc = 7

            option = MyOption()

        x = []

        def MyGreenPool(**kw):
            x.append(kw)
            # Building a Detox object will already call GreenPool(),
            # so we have to let MyGreenPool being called twice before raise
            if len(x) == 2:
                raise ValueError

        from detox import proc

        setattr(proc, "GreenPool", MyGreenPool)
        with pytest.raises(ValueError):
            d = proc.Detox(MyConfig())
            d.runtestsmulti(["env1", "env2", "env3"])  # Fake env list

        assert x[0] == {}  # When building Detox object
        assert x[1] == {"size": 7}  # When calling runtestsmulti

    @pytest.mark.timeout(60)
    def test_runtests(self, cmd):
        now1 = datetime.now()
        cmd.rundetox("-n", "1", "-epy1,py2")
        then1 = datetime.now()
        delta1 = then1 - now1
        assert delta1 >= timedelta(seconds=2)

        now2 = datetime.now()
        cmd.rundetox("--num", "2", "-epy1,py2")
        then2 = datetime.now()
        delta2 = then2 - now2
        assert delta2 >= timedelta(seconds=1)

        assert delta1 >= delta2, "pool size=2 took much time than pool size=1"
